import dataclasses as dc
import datetime
import threading
import typing as t
from enum import Enum

from pymongo.change_stream import CollectionChangeStream

from aladindb import logging
from aladindb.backends.mongodb import query
from aladindb.cdc import cdc
from aladindb.misc.runnable.runnable import Event

from .base import CachedTokens, TokenType

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer

MongoChangePipelines: t.Dict[str, t.Sequence[t.Any]] = {'generic': []}


class CDCKeys(str, Enum):
    """
    A enum to represent mongo change document keys.
    """

    operation_type = 'operationType'
    document_key = 'documentKey'
    update_descriptions_key = 'updateDescription'
    update_field_key = 'updatedFields'
    document_data_key = 'fullDocument'
    deleted_document_data_key = 'documentKey'


_CDCKEY_MAP = {
    cdc.DBEvent.update: CDCKeys.document_key,
    cdc.DBEvent.insert: CDCKeys.document_data_key,
    cdc.DBEvent.delete: CDCKeys.deleted_document_data_key,
}


@dc.dataclass
class MongoChangePipeline:
    """`MongoChangePipeline` is a class to represent listen pipeline
    in mongodb watch api.
    """

    matching_operations: t.Sequence[str] = dc.field(default_factory=list)

    def validate(self):
        raise NotImplementedError

    def build_matching(self) -> t.Sequence[t.Dict]:
        """A helper fxn to build a listen pipeline for mongo watch api.

        :param matching_operations: A list of operations to watch.
        """
        if bad := [op for op in self.matching_operations if op not in cdc.DBEvent]:
            raise ValueError(f'Unknown operations: {bad}')

        return [{'$match': {'operationType': {'$in': [*self.matching_operations]}}}]


class MongoDatabaseListener(cdc.BaseDatabaseListener):
    """
    It is a class which helps capture data from mongodb database and handle it
    accordingly.

    This class accepts options and db instance from user and starts a scheduler
    which could schedule a listening service to listen change stream.

    This class builds a workflow graph on each change observed.

    """

    DEFAULT_ID: str = '_id'
    EXCLUSION_KEYS: t.Sequence[str] = [DEFAULT_ID]
    IDENTITY_SEP: str = '/'
    _scheduler: t.Optional[threading.Thread]

    _change_pipeline: t.Union[str, t.Sequence[t.Dict], None] = None

    def __init__(
        self,
        db: 'Datalayer',
        on: query.Collection,
        stop_event: Event,
        identifier: 'str' = '',
        timeout: t.Optional[float] = None,
        resume_token: t.Optional[TokenType] = None,
    ):
        """__init__.

        :param db: It is a aladindb instance.
        :param on: It is used to define a Collection on which CDC would be performed.
        :param stop_event: A threading event flag to notify for stoppage.
        :param identifier: A identifier to represent the listener service.
        :param resume_token: A resume token is a token used to resume
        the change stream in mongo.
        """

        self.tokens = CachedTokens()
        self.resume_token = None

        if resume_token is not None:
            self.resume_token = resume_token

        self._change_pipeline = None

        super().__init__(
            db=db, on=on, stop_event=stop_event, identifier=identifier, timeout=timeout
        )

    def on_create(
        self, ids: t.Sequence, db: 'Datalayer', collection: query.Collection
    ) -> None:
        """on_create.
        A helper on create event handler which handles inserted document in the
        change stream.
        It basically extracts the change document and build the taskflow graph to
        execute.

        :param ids: Changed document ids.
        :param db: a aladindb instance.
        :param collection: The collection on which change was observed.
        """
        logging.debug('Triggered `on_create` handler.')
        # new document added!
        self.create_event(
            ids=ids, db=db, table_or_collection=collection, event=cdc.DBEvent.insert
        )

    def on_update(
        self, ids: t.Sequence, db: 'Datalayer', collection: query.Collection
    ) -> None:
        """on_update.

        A helper on update event handler which handles updated document in the
        change stream.
        It basically extracts the change document and build the taskflow graph to
        execute.

        :param ids: Changed document ids.
        :param db: a aladindb instance.
        :param collection: The collection on which change was observed.
        """
        logging.debug('Triggered `on_update` handler.')
        self.create_event(
            ids=ids, db=db, table_or_collection=collection, event=cdc.DBEvent.insert
        )

    def on_delete(
        self, ids: t.Sequence, db: 'Datalayer', collection: query.Collection
    ) -> None:
        """on_delete.

        A helper on delete event handler which handles deleted document in the
        change stream.
        It basically extracts the change document and build the taskflow graph to
        execute.

        :param ids: Changed document ids.
        :param db: a aladindb instance.
        :param collection: The collection on which change was observed.
        """
        logging.debug('Triggered `on_delete` handler.')
        self.create_event(
            ids=ids, db=db, table_or_collection=collection, event=cdc.DBEvent.delete
        )

    @staticmethod
    def _get_stream_pipeline(
        change: str,
    ) -> t.Optional[t.Sequence[t.Any]]:
        """_get_stream_pipeline.

        :param change: change can be a prebuilt listen pipeline like
        'generic' or user Defined listen pipeline.
        """
        return MongoChangePipelines.get(change)

    def _get_reference_id(self, document: t.Dict) -> t.Optional[str]:
        """_get_reference_id.

        :param document:
        """
        try:
            document_key = document[CDCKeys.document_key]
            reference_id = str(document_key['_id'])
        except KeyError:
            return None
        return reference_id

    def dump_token(self, change: t.Dict) -> None:
        """dump_token.
        A helper utility to dump resume token from the changed document.

        :param change:
        """
        token = change[self.DEFAULT_ID]
        self.tokens.append(token)

    def check_if_taskgraph_change(self, change: t.Dict) -> bool:
        """
        A helper method to check if the cdc change is done
        by taskgraph nodes.
        """
        if change[CDCKeys.operation_type] == cdc.DBEvent.update:
            updates = change[CDCKeys.update_descriptions_key]
            updated_fields = updates[CDCKeys.update_field_key]
            return any(
                [True for k in updated_fields.keys() if k.startswith('_outputs')]
            )
        return False

    def setup_cdc(self) -> CollectionChangeStream:
        """
        Setup cdc change stream from user provided
        """
        if isinstance(self._change_pipeline, str):
            pipeline = self._get_stream_pipeline(self._change_pipeline)

        elif isinstance(self._change_pipeline, list):
            pipeline = self._change_pipeline or None

        else:
            raise TypeError(
                'Change pipeline can be either a string or a dictionary, '
                f'provided {type(self._change_pipeline)}'
            )
        stream = self._on_component.change_stream(
            pipeline=pipeline, resume_after=self.resume_token
        )

        stream_iterator = stream(self.db)

        logging.info("Started listening database with identity ", self.identity)

        return stream_iterator

    def next_cdc(self, stream: CollectionChangeStream) -> None:
        """
        Get the next stream of change observed on the given `Collection`.
        """

        change = stream.try_next()
        if change is not None:
            logging.debug(f'Database change encountered at {datetime.datetime.now()}')

            if not self.check_if_taskgraph_change(change):
                reference_id = self._get_reference_id(change)

                if not reference_id:
                    logging.warn('Document change not handled due to no document key')
                    return

                event = change[CDCKeys.operation_type]
                ids = [change[_CDCKEY_MAP[event]][self.DEFAULT_ID]]
                self.event_handler(ids, event)
            self.dump_token(change)

    def set_change_pipeline(
        self, change_pipeline: t.Optional[t.Union[str, t.Sequence[t.Dict]]]
    ) -> None:
        """
        Set the change pipeline for the listener.
        """
        if change_pipeline is None:
            change_pipeline = MongoChangePipelines.get('generic')

        self._change_pipeline = change_pipeline

    def listen(
        self,
        change_pipeline: t.Optional[t.Union[str, t.Sequence[t.Dict]]] = None,
    ) -> None:
        """Primary fxn to initiate listening of a database on the collection
        with defined `change_pipeline` by the user.

        :param change_pipeline: A mongo listen pipeline defined by the user
        for more fine grained listening.
        """
        try:
            self._stop_event.clear()
            if self._scheduler:
                if self._scheduler.is_alive():
                    raise RuntimeError(
                        'CDC Listener thread is already running!,/'
                        'Please stop the listener first.'
                    )

            self._scheduler = cdc.DatabaseListenerThreadScheduler(
                self, stop_event=self._stop_event, start_event=self._startup_event
            )
            self.set_change_pipeline(change_pipeline)

            assert self._scheduler is not None
            self._scheduler.start()

            self._startup_event.wait(timeout=self.timeout)
        except Exception:
            logging.error('Listening service stopped!')
            self.stop()
            raise

    def resume_tokens(self) -> t.Sequence[TokenType]:
        """
        Get the resume tokens from the change stream.
        """
        return self.tokens.load()

    def stop(self) -> None:
        """
        Stop listening cdc changes.
        This stops the corresponding services as well.
        """
        self._stop_event.set()
        if self._scheduler:
            self._scheduler.join()

    def running(self) -> bool:
        return not self._stop_event.is_set()



import dataclasses as dc
import typing as t

import cohere
import tqdm
from cohere.error import CohereAPIError, CohereConnectionError

from aladindb.backends.ibis.data_backend import IbisDataBackend
from aladindb.backends.ibis.field_types import dtype
from aladindb.base.datalayer import Datalayer
from aladindb.components.model import APIModel
from aladindb.components.vector_index import sqlvector, vector
from aladindb.ext.utils import format_prompt, get_key
from aladindb.misc.retry import Retry

retry = Retry(exception_types=(CohereAPIError, CohereConnectionError))

KEY_NAME = 'COHERE_API_KEY'


@dc.dataclass(kw_only=True)
class Cohere(APIModel):
    """Cohere predictor"""

    client_kwargs: t.Dict[str, t.Any] = dc.field(default_factory=dict)

    def __post_init__(self):
        super().__post_init__()
        self.identifier = self.identifier or self.model


@dc.dataclass(kw_only=True)
class CohereEmbed(Cohere):
    """Cohere embedding predictor

    :param shape: The shape as ``tuple`` of the embedding.
    """

    shapes: t.ClassVar[t.Dict] = {'embed-english-v2.0': (4096,)}
    shape: t.Optional[t.Sequence[int]] = None

    def __post_init__(self):
        super().__post_init__()
        if self.shape is None:
            self.shape = self.shapes[self.identifier]

    def pre_create(self, db):
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend):
            if self.encoder is None:
                self.encoder = sqlvector(self.shape)
        elif self.encoder is None:
            self.encoder = vector(self.shape)

    @retry
    def _predict_one(self, X: str, **kwargs):
        client = cohere.Client(get_key(KEY_NAME), **self.client_kwargs)
        e = client.embed(texts=[X], model=self.identifier, **kwargs)
        return e.embeddings[0]

    @retry
    async def _apredict_one(self, X: str, **kwargs):
        client = cohere.AsyncClient(get_key(KEY_NAME), **self.client_kwargs)
        e = await client.embed(texts=[X], model=self.identifier, **kwargs)
        await client.close()
        return e.embeddings[0]

    @retry
    def _predict_a_batch(self, texts: t.List[str], **kwargs):
        client = cohere.Client(get_key(KEY_NAME), **self.client_kwargs)
        out = client.embed(texts=texts, model=self.identifier, **kwargs)
        return [r for r in out.embeddings]

    @retry
    async def _apredict_a_batch(self, texts: t.List[str], **kwargs):
        client = cohere.AsyncClient(get_key(KEY_NAME), **self.client_kwargs)
        out = await client.embed(texts=texts, model=self.identifier, **kwargs)
        await client.close()
        return [r for r in out.embeddings]

    def _predict(self, X, one=False, **kwargs):
        if isinstance(X, str):
            return self._predict_one(X)
        out = []
        batch_size = kwargs.pop('batch_size', 100)
        for i in tqdm.tqdm(range(0, len(X), batch_size)):
            out.extend(self._predict_a_batch(X[i : i + batch_size], **kwargs))
        return out

    async def _apredict(self, X, one=False, **kwargs):
        if isinstance(X, str):
            return await self._apredict_one(X)
        out = []
        batch_size = kwargs.pop('batch_size', 100)
        for i in range(0, len(X), batch_size):
            out.extend(await self._apredict_a_batch(X[i : i + batch_size], **kwargs))
        return out


@dc.dataclass(kw_only=True)
class CohereGenerate(Cohere):
    """Cohere realistic text generator (chat predictor)

    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to use to seed the response.
    """

    takes_context: bool = True
    prompt: str = ''

    def pre_create(self, db: Datalayer) -> None:
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend) and self.encoder is None:
            self.encoder = dtype('str')

    @retry
    def _predict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = format_prompt(X, self.prompt, context=context)
        client = cohere.Client(get_key(KEY_NAME), **self.client_kwargs)
        resp = client.generate(prompt=X, model=self.identifier, **kwargs)
        return resp.generations[0].text

    @retry
    async def _apredict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = format_prompt(X, self.prompt, context=context)
        client = cohere.AsyncClient(get_key(KEY_NAME), **self.client_kwargs)
        resp = await client.generate(prompt=X, model=self.identifier, **kwargs)
        await client.close()
        return resp.generations[0].text

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        return [self._predict_one(msg) for msg in X]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        return [await self._apredict_one(msg) for msg in X]



import dataclasses as dc

from aladindb.ext.llm.base import BaseOpenAI


@dc.dataclass
class OpenAI(BaseOpenAI):
    """
    OpenAI chat completion predictor.
    {parent_doc}
    """

    __doc__ = __doc__.format(parent_doc=BaseOpenAI.__doc__)

    def __post_init__(self):
        """Set model name."""
        # only support chat mode
        self.chat = True
        super().__post_init__()



from __future__ import annotations

import typing as t
from contextlib import contextmanager

if t.TYPE_CHECKING:
    from torch import device as _device
    from torch.nn.modules import Module


def device_of(module: Module) -> t.Union[_device, str]:
    """
    Get device of a model.

    :param model: PyTorch model
    """
    try:
        return next(iter(module.state_dict().values())).device
    except StopIteration:
        return 'cpu'


@contextmanager
def eval(module: Module) -> t.Iterator[None]:
    """
    Temporarily set a module to evaluation mode.

    :param module: PyTorch module
    """
    was_training = module.training
    try:
        module.eval()
        yield
    finally:
        if was_training:
            module.train()


@contextmanager
def set_device(module: Module, device: _device):
    """
    Temporarily set a device of a module.

    :param module: PyTorch module
    :param device: Device to set
    """
    device_before = device_of(module)
    try:
        module.to(device)
        yield
    finally:
        module.to(device_before)


def to_device(
    item: t.Any,  # lists or dicts of Tensors
    device: t.Union[str, _device],
) -> t.Any:
    """
    Send tensor leaves of nested list/ dictionaries/ tensors to device.

    :param item: torch.Tensor instance
    :param device: device to which one would like to send
    """
    if isinstance(item, tuple):
        item = list(item)
    if isinstance(item, list):
        for i, it in enumerate(item):
            item[i] = to_device(it, device)
        return item
    if isinstance(item, dict):
        for k in item:
            item[k] = to_device(item[k], device)
        return item
    if hasattr(item, 'to'):
        return item.to(device)
    return item



import dataclasses as dc
import typing as t

import anthropic
from anthropic import APIConnectionError, APIError, APIStatusError, APITimeoutError

from aladindb.backends.ibis.data_backend import IbisDataBackend
from aladindb.backends.ibis.field_types import dtype
from aladindb.base.datalayer import Datalayer
from aladindb.components.model import APIModel
from aladindb.ext.utils import format_prompt, get_key
from aladindb.misc.retry import Retry

retry = Retry(
    exception_types=(APIConnectionError, APIError, APIStatusError, APITimeoutError)
)

KEY_NAME = 'ANTHROPIC_API_KEY'


@dc.dataclass(kw_only=True)
class Anthropic(APIModel):
    """Anthropic predictor."""

    client_kwargs: t.Dict[str, t.Any] = dc.field(default_factory=dict)

    def __post_init__(self):
        self.identifier = self.identifier or self.model


@dc.dataclass(kw_only=True)
class AnthropicCompletions(Anthropic):
    """Cohere completions (chat) predictor.

    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to use to seed the response.
    """

    prompt: str = ''

    def pre_create(self, db: Datalayer) -> None:
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend) and self.encoder is None:
            self.encoder = dtype('str')

    @retry
    def _predict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = format_prompt(X, self.prompt, context=context)
        client = anthropic.Anthropic(api_key=get_key(KEY_NAME), **self.client_kwargs)
        resp = client.completions.create(prompt=X, model=self.identifier, **kwargs)
        return resp.completion

    @retry
    async def _apredict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = format_prompt(X, self.prompt, context=context)
        client = anthropic.AsyncAnthropic(
            api_key=get_key(KEY_NAME), **self.client_kwargs
        )
        resp = await client.completions.create(
            prompt=X, model=self.identifier, **kwargs
        )
        return resp.completion

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        return [self._predict_one(msg) for msg in X]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        return [await self._apredict_one(msg) for msg in X]



import dataclasses as dc
import typing as t

import tqdm

from aladindb.backends.ibis.data_backend import IbisDataBackend
from aladindb.components.model import APIModel
from aladindb.components.vector_index import sqlvector, vector
from aladindb.ext.jina.client import JinaAPIClient


@dc.dataclass(kw_only=True)
class Jina(APIModel):
    """Cohere predictor"""

    api_key: t.Optional[str] = None

    def __post_init__(self):
        super().__post_init__()
        self.identifier = self.identifier or self.model
        self.client = JinaAPIClient(model_name=self.identifier, api_key=self.api_key)


@dc.dataclass(kw_only=True)
class JinaEmbedding(Jina):
    """Jina embedding predictor

    :param shape: The shape of the embedding as ``tuple``.
        If not provided, it will be obtained by sending a simple query to the API
    """

    shape: t.Optional[t.Sequence[int]] = None

    def __post_init__(self):
        super().__post_init__()
        if self.shape is None:
            self.shape = (len(self.client.encode_batch(['shape'])[0]),)

    def pre_create(self, db):
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend):
            if self.encoder is None:
                self.encoder = sqlvector(self.shape)
        elif self.encoder is None:
            self.encoder = vector(self.shape)

    def _predict_one(self, X: str, **kwargs):
        return self.client.encode_batch([X])[0]

    async def _apredict_one(self, X: str, **kwargs):
        embeddings = await self.client.aencode_batch([X])
        return embeddings[0]

    def _predict_a_batch(self, texts: t.List[str], **kwargs):
        return self.client.encode_batch(texts)

    async def _apredict_a_batch(self, texts: t.List[str], **kwargs):
        return await self.client.aencode_batch(texts)

    def _predict(self, X, one=False, **kwargs):
        if isinstance(X, str):
            return self._predict_one(X)
        out = []
        batch_size = kwargs.pop('batch_size', 100)
        for i in tqdm.tqdm(range(0, len(X), batch_size)):
            out.extend(self._predict_a_batch(X[i : i + batch_size], **kwargs))
        return out

    async def _apredict(self, X, one=False, **kwargs):
        if isinstance(X, str):
            return await self._apredict_one(X)
        out = []
        batch_size = kwargs.pop('batch_size', 100)
        for i in range(0, len(X), batch_size):
            out.extend(await self._apredict_a_batch(X[i : i + batch_size], **kwargs))
        return out



import typing as t

from dask import distributed

from aladindb import logging
from aladindb.backends.base.compute import ComputeBackend


class DaskComputeBackend(ComputeBackend):
    """
    A client for interacting with a Dask cluster. Initialize the DaskClient.

    :param address: The address of the Dask cluster.
    :param serializers: A list of serializers to be used by the client. (optional)
    :param deserializers: A list of deserializers to be used by the client. (optional)
    :param local: Set to True to create a local Dask cluster. (optional)
    :param **kwargs: Additional keyword arguments to be passed to the DaskClient.
    """

    def __init__(
        self,
        address: t.Optional[str] = None,
        serializers: t.Optional[t.Sequence[t.Callable]] = None,
        deserializers: t.Optional[t.Sequence[t.Callable]] = None,
        local: bool = False,
        **kwargs,
    ):
        self._futures_collection: t.Dict[str, distributed.Future] = {}

        if local:
            # Create and connect to the local cluster.
            self.client = distributed.Client(processes=False)
        else:
            assert address, 'Address cannot be ``None`` for non local dask client'
            # Connect to a remote cluster.
            self.client = distributed.Client(
                address=address,
                serializers=serializers,
                deserializers=deserializers,
                **kwargs,
            )

    @property
    def type(self) -> str:
        return "distributed"

    @property
    def name(self) -> str:
        return "dask"

    def submit(self, function: t.Callable, *args, **kwargs) -> distributed.Future:
        """
        Submits a function to the Dask server for execution.

        :param function: The function to be executed.
        """
        future = self.client.submit(function, *args, **kwargs)
        self._futures_collection[future.key] = future

        logging.success(f"Job submitted.  function:{function} future:{future}")
        return future

    @property
    def tasks(self) -> t.Dict[str, distributed.Future]:
        """
        List for all pending tasks
        """
        return self._futures_collection

    def wait(self, identifier: str) -> None:
        """
        Waits for all pending tasks to complete.
        :param identifier: Future task reference
        """
        futures = [self._futures_collection.get(identifier)]
        distributed.wait(futures)

    def wait_all(self) -> None:
        """
        Waits for all pending tasks to complete.
        """
        futures = list(self._futures_collection.values())
        distributed.wait(futures)

    def result(self, identifier: str) -> t.Any:
        """
        Retrieves the result of a previously submitted task.
        Note: This will block until the future is completed.

        :param identifier: The identifier of the submitted task.
        """
        future = self._futures_collection[identifier]
        return self.client.gather(future)

    def disconnect(self) -> None:
        """
        Disconnect the Dask client.
        """
        self.client.close()

    def shutdown(self) -> None:
        """
        Shuts down the Dask cluster.
        """
        self.client.shutdown()
